use crate::webutil::internal_error;
use axum::body::{boxed, Bytes};
use axum::extract::{Extension, Query};
use axum::http::StatusCode;
use axum::response::{IntoResponse, Response};
use quickpeep_index::auxiliary::icon_store::IconStore;
use serde::Deserialize;
use std::borrow::Borrow;
use std::sync::Arc;

#[derive(Deserialize, Debug, Clone)]
pub struct IconRetrievalParameters {
    /// The search query
    b: String,
}

pub async fn retrieve_icon(
    icon_store: Extension<Arc<IconStore>>,
    params: Query<IconRetrievalParameters>,
) -> impl IntoResponse {
    retrieve_icon_inner(icon_store, params)
        .await
        .map_err(internal_error)
}

pub async fn retrieve_icon_inner(
    Extension(icon_store): Extension<Arc<IconStore>>,
    Query(params): Query<IconRetrievalParameters>,
) -> anyhow::Result<Response> {
    let favicon_url_hash = u64::from_str_radix(&params.b, 16)?;
    let favicon_url_hash: [u8; 8] = favicon_url_hash.to_le_bytes();
    let txn = icon_store.env.ro_txn()?;
    match icon_store.icons.get(&txn, &favicon_url_hash)? {
        None => Ok((StatusCode::NOT_FOUND, "Not found").into_response()),
        Some(icon_bytes) => Ok(Response::builder()
            .header("content-type", "image/webp")
            .body(boxed(
                Bytes::copy_from_slice(icon_bytes.borrow()).into_response(),
            ))?),
    }
}
