use crate::steps::{bom_sniff, prescan, BOM_SNIFF_NEEDED_BYTES};
use chardetng::EncodingDetector;
use encoding_rs::Encoding;

pub mod steps;

/// The spec requires document authors to place their <meta> tags in the first 1024 bytes.
pub const SNIFF_WINDOW_SIZE: usize = 1024;

/// Attempts to implement the 'certain' stages of the encoding sniffing algorithm described at:
/// https://html.spec.whatwg.org/multipage/parsing.html#determining-the-character-encoding
///
/// You should pass in at least the first 3 bytes as the `sniff_window` here.
pub fn sniff_with_certain_confidence(
    sniff_window: &[u8],
    content_type_header: Option<&[u8]>,
) -> Option<&'static Encoding> {
    // 1. BOM sniffing.
    if sniff_window.len() > BOM_SNIFF_NEEDED_BYTES {
        if let Some(encoding) = bom_sniff(
            sniff_window[0..BOM_SNIFF_NEEDED_BYTES]
                .try_into()
                .expect("checked size cast"),
        ) {
            return Some(encoding);
        }
    }

    // 2. User override (Not implemented)

    // 3. 'Wait for bytes' — we already have 1024

    // 4. If the transport layer specifies an encoding, return as certain.
    extract_encoding_from_content_type_header(content_type_header.unwrap_or(b""))
}

/// Implementing the all the stages of the encoding sniffing algorithm described at:
/// https://html.spec.whatwg.org/multipage/parsing.html#determining-the-character-encoding
/// except where those stages rely on other information (e.g. 'prior knowledge', a user default
/// or a parent window context).
///
/// You should pass in at least the first 1024 bytes as the `sniff_window` here.
pub fn sniff(
    sniff_window: &[u8],
    is_sniff_window_the_entire_document: bool,
    content_type_header: Option<&[u8]>,
) -> &'static Encoding {
    if let Some(certain) = sniff_with_certain_confidence(sniff_window, content_type_header) {
        return certain;
    }

    // 5. Optionally prescan the byte stream to determine its encoding
    if let Some(prescan_tentative) = prescan(sniff_window) {
        return prescan_tentative;
    }

    // 8. The user agent may attempt to autodetect the character encoding from applying frequency
    // analysis or other algorithms to the data stream. Such algorithms may use information about
    // the resource other than the resource's contents, including the address of the resource.
    // If autodetection succeeds in determining a character encoding, and that encoding is a
    // supported encoding, then return that encoding, with the confidence tentative.
    let mut detector = EncodingDetector::new();
    detector.feed(sniff_window, is_sniff_window_the_entire_document);
    // 'Allow UTF-8' should be set to false for non-file: URIs, apparently.
    // We do that here, though I'm not sure this is what we actually want outside of a browser...
    detector.guess(None, false)
}

pub fn extract_encoding_from_content_type_header(
    content_type_header: &[u8],
) -> Option<&'static Encoding> {
    for header_part in content_type_header.split(|b| *b == b';') {
        // To-UTF-8-lossy is definitely not spec compliant, but trim_ascii() on byte slices is
        // unstable, so let's just help ourselves out of a pickle.
        let header_part_ascii_ish = String::from_utf8_lossy(header_part);
        let key_value: Vec<&str> = header_part_ascii_ish.trim().split("=").collect();
        let key = key_value.get(0).cloned().unwrap_or("");
        if key.to_ascii_lowercase() == "charset" {
            let value = key_value.get(1).cloned().unwrap_or("");
            return Encoding::for_label(value.as_bytes());
        }
    }
    None
}

#[cfg(test)]
mod test {
    use crate::sniff;

    #[test]
    fn test_simple_cases() {
        assert_eq!(sniff(b"<u>hi</u>", true, None).name(), "windows-1252");

        assert_eq!(
            sniff(b"<meta charset=UTF8><u>hi</u>", true, None).name(),
            "UTF-8"
        );

        assert_eq!(
            sniff(
                b"<meta charset=UTF8><u>hi</u>",
                true,
                Some("text/html; charset=Shift-JIS".as_bytes())
            )
            .name(),
            "Shift_JIS"
        );

        assert_eq!(
            sniff(b"<!-- haha we wish <meta charset=UTF8> --><meta http-equiv='content-type' content='text/html; charset=Shift-JIS'><u>hi</u>", true, None).name(),
            "Shift_JIS"
        );
    }
}
