use crate::backend::{Backend, BackendIndependentDocument, SearchDocument};
use anyhow::Context;
use fancy_mdbx::database::WrappedTable;
use fancy_mdbx::environment::Env;
use fancy_mdbx::wrapper::{CompressorWrapper, SerdeBareWrapper};
use serde::{Deserialize, Serialize};
use std::collections::HashMap;
use std::path::Path;
use tantivy::collector::TopDocs;
use tantivy::query::QueryParser;
use tantivy::schema::{Facet, Field, SchemaBuilder, STORED, TEXT};
use tantivy::tokenizer::TokenizerManager;
use tantivy::{doc, Index, IndexWriter, SnippetGenerator};

pub struct Fields {
    title: Field,
    article: Field,
    nonarticle: Field,
    url: Field,
    tags: Field,
}

pub struct TantivyBackend {
    index: Index,
    fields: Fields,
    index_writer: Option<IndexWriter>,
    env: Env,
    tables: StoreTables,
}

#[derive(Serialize, Deserialize, Debug, Clone)]
pub struct DocumentStoreRow {
    title: String,
    body: String,
    nonbody: String,
    /// BLAKE3 of the favicon URL
    favicon_url_hash: u64,
}

pub struct StoreTables {
    pub documents: WrappedTable<String, CompressorWrapper<SerdeBareWrapper<DocumentStoreRow>>>,
}

impl TantivyBackend {
    pub fn open(path: &Path) -> anyhow::Result<TantivyBackend> {
        if !path.exists() {
            std::fs::create_dir(path)?;
        }

        let dir_path = path.join("tantivy");

        let (index, fields) = if dir_path.exists() {
            let index = Index::open_in_dir(&dir_path).context("failed to open index")?;

            let schema = index.schema();
            let mut field_map: HashMap<_, _> = schema
                .fields()
                .map(|(field, field_entry)| (field_entry.name(), field))
                .collect();

            let fields = Fields {
                title: field_map.remove("title").context("No title field")?,
                article: field_map.remove("article").context("No article field")?,
                nonarticle: field_map
                    .remove("nonarticle")
                    .context("No nonarticle field")?,
                url: field_map.remove("url").context("No url field")?,
                tags: field_map.remove("tags").context("No tags field")?,
            };

            (index, fields)
        } else {
            let mut schema_builder = SchemaBuilder::new();
            let fields = Fields {
                title: schema_builder.add_text_field("title", TEXT),
                article: schema_builder.add_text_field("article", TEXT),
                nonarticle: schema_builder.add_text_field("nonarticle", TEXT),
                url: schema_builder.add_text_field("url", STORED),
                tags: schema_builder.add_facet_field("tags", ()),
            };
            let schema = schema_builder.build();

            std::fs::create_dir(&dir_path)?;
            let index =
                Index::create_in_dir(&dir_path, schema).context("failed to create index")?;

            (index, fields)
        };

        // Open a libMDBX environment
        let env = Env::open(&dir_path.join("store.mdbx"))?;

        // TODO use a dictionary for the compression
        let tables = StoreTables {
            documents: env.open_wrapped_table(
                Some("documents"),
                (),
                Default::default(),
                CompressorWrapper::new()?,
            )?,
        };

        Ok(TantivyBackend {
            index,
            fields,
            index_writer: None,
            env,
            tables,
        })
    }
}

impl Backend for TantivyBackend {
    fn add_document(&mut self, document: BackendIndependentDocument) -> anyhow::Result<()> {
        let index_writer = match self.index_writer.as_ref() {
            None => {
                self.index_writer = Some(self.index.writer(100 * 1024 * 1024)?);
                self.index_writer.as_ref().unwrap()
            }
            Some(index_writer) => index_writer,
        };

        let Fields {
            title,
            article,
            nonarticle,
            url,
            tags,
        } = &self.fields;

        let mut tantivy_doc = doc! {
            *title => document.title.clone(),
            *article => document.article_body.clone(),
            *nonarticle => document.nonarticle_body.clone(),
            *url => document.url.clone(),
        };
        // TODO do we actually want facets? How about u64 tags or something...?
        for tag in &document.tags {
            tantivy_doc.add_facet(*tags, Facet::from(&format!("/{}", tag)));
        }

        index_writer.add_document(tantivy_doc)?;
        self.env.rw_txn(|txn| {
            self.tables.documents.put(
                txn,
                &document.url,
                &DocumentStoreRow {
                    title: document.title,
                    body: document.article_body,
                    nonbody: document.nonarticle_body,
                    favicon_url_hash: document.favicon_url_hash,
                },
            )
        })?;
        Ok(())
    }

    fn flush(&mut self) -> anyhow::Result<()> {
        if let Some(index_writer) = self.index_writer.as_mut() {
            index_writer.commit()?;
        }
        Ok(())
    }

    fn query(&self, query: String) -> anyhow::Result<Vec<SearchDocument>> {
        let reader = self.index.reader()?;
        let parser = QueryParser::new(
            self.index.schema(),
            vec![
                self.fields.title,
                self.fields.article,
                self.fields.nonarticle,
            ],
            TokenizerManager::default(),
        );

        let query = parser.parse_query(&query)?;

        // TODO tweak scores?
        let txn = self.env.ro_txn()?;
        let searcher = reader.searcher();
        let results = searcher.search(&query, &TopDocs::with_limit(50))?;
        let mut out = Vec::with_capacity(results.len());

        let article_snippet_generator =
            SnippetGenerator::create(&searcher, &query, self.fields.article)?;
        // TODO ? let nonarticle_snippet_generator = SnippetGenerator::create(&searcher, &query, self.fields.article)?;

        for (score, doc_address) in results {
            let doc = searcher.doc(doc_address)?;
            let url = doc
                .get_first(self.fields.url)
                .context("No URL field")?
                .as_text()
                .context("URL not text")?;
            let doc_row = self
                .tables
                .documents
                .get(&txn, url.to_owned())?
                .context("Document row not found in doc store")?;

            let snippet = article_snippet_generator.snippet(&doc_row.body);
            let excerpt = snippet.to_html();

            let tags = doc
                .get_all(self.fields.tags)
                .map(|fv| {
                    String::from(
                        *fv.as_facet()
                            .expect("tags must be facet!")
                            .to_path()
                            .last()
                            .unwrap_or(&""),
                    )
                })
                .collect();

            out.push(SearchDocument {
                score,
                title: doc_row.title,
                excerpt,
                tags,
                url: url.to_owned(),
                favicon_url_hash: doc_row.favicon_url_hash,
            });
        }

        Ok(out)
    }
}
