use anyhow::{bail, Context};
use clap::Parser;
use colour::{blue, yellow_ln};
use env_logger::Env;
use std::collections::{BTreeSet, HashMap};
use std::fs::{File, OpenOptions};
use std::io::{BufRead, BufReader, Write};

use patricia_tree::PatriciaMap;
use quickpeep_densedoc::DenseTree;
use quickpeep_index::auxiliary::icon_store::IconStore;
use quickpeep_index::backend::{Backend, BackendIndependentDocument};
use quickpeep_indexer::config::IndexerConfig;
use quickpeep_seed_parser::loader::{
    find_seed_files, seed_loader, Seed, UrlOrUrlPattern, SEED_EXTENSION,
};
use quickpeep_structs::rake_entries::{
    AnalysisAntifeatures, IconEntry, PackRecord, RakedPageEntry, SCHEMA_RAKED_ICONS,
    SCHEMA_RAKED_PAGES,
};
use quickpeep_utils::urls::get_reduced_domain;
use smartstring::alias::CompactString;
use std::path::{Path, PathBuf};
use tempfile::NamedTempFile;
use tokio::sync::mpsc::Receiver;
use url::Url;

/// Seeds a raker's queue with URLs
#[derive(Clone, Debug, Parser)]
pub struct Opts {
    #[clap(long = "config")]
    config: Option<PathBuf>,

    /// If specified, rakepacks from a feed will automatically be fetched and indexed.
    /// The rakepacks are tracked as having been processed.
    #[clap(long = "feed")]
    feed: Option<Url>,

    rakepacks: Vec<PathBuf>,
}

#[tokio::main]
pub async fn main() -> anyhow::Result<()> {
    env_logger::Builder::from_env(Env::default().default_filter_or("info,qp_indexer=debug")).init();

    let opts: Opts = Opts::parse();

    let config_path = opts
        .config
        .unwrap_or_else(|| PathBuf::from("quickpeep.ron"));
    let config = IndexerConfig::load(&config_path).context("Failed to load config")?;

    let icon_store =
        IconStore::open(config.index.icon_store.as_path()).context("failed to open icon store")?;

    let seed_files = find_seed_files(config.seed_dir.clone(), SEED_EXTENSION)
        .await
        .context("failed to find seed files")?;
    let (seed_tx, seed_rx) = tokio::sync::mpsc::channel(64);
    let handle = tokio::spawn(async move {
        seed_loader(seed_files, &seed_tx).await?;
        Ok(()) as anyhow::Result<()>
    });

    let seed_lookup = build_seed_lookup_table(seed_rx).await?;

    handle.await??;

    let mut indexer_backend = config.open_indexer_backend()?;

    if let Some(feed) = opts.feed {
        let processed_rakepack_path = config
            .processed_rakepack_path()
            .context("can't get a suitable location to track processed rakepacks")?;
        handle_pack_feed(
            feed,
            &mut indexer_backend,
            processed_rakepack_path,
            &seed_lookup,
            &icon_store,
        )
        .context("failed to handle pack feed")?;
    }

    for pack in opts.rakepacks {
        handle_pack(&pack, &mut indexer_backend, &seed_lookup, &icon_store)
            .with_context(|| format!("Whilst handling pack: {pack:?}"))?;
    }
    indexer_backend.flush()?;

    Ok(())
}

pub fn handle_pack_feed(
    feed_url: Url,
    indexer_backend: &mut Box<dyn Backend>,
    processed_list_path: PathBuf,
    seed_lookup: &SeedLookupTable,
    icon_store: &IconStore,
) -> anyhow::Result<()> {
    blue!("Scanning feed: ");
    yellow_ln!("{:?}", feed_url);

    let new_packs =
        find_new_packs(feed_url.clone(), &processed_list_path).context("finding new packs")?;
    let mut processed_log = OpenOptions::new()
        .append(true)
        .create(true)
        .open(&processed_list_path)
        .context("can't open processed list for append")?;

    for pack_name in new_packs {
        let pack_url = feed_url
            .join(&pack_name)
            .context("Can't resolve URL of new pack")?;

        blue!("Downloading: ");
        yellow_ln!("{:?}", pack_url);
        let mut temp_file = NamedTempFile::new().context("opening temp file")?;

        reqwest::blocking::get(pack_url.clone())
            .context("failed to request pack")?
            .error_for_status()?
            .copy_to(temp_file.as_file_mut())
            .context("failed to download pack to temp file")?;

        handle_pack(temp_file.path(), indexer_backend, seed_lookup, icon_store).with_context(
            || {
                format!(
                    "Whilst handling pack: {:?} ({:?})",
                    temp_file.path(),
                    pack_url
                )
            },
        )?;

        processed_log.write(format!("\n{}", &pack_name).as_bytes())?;
        processed_log.flush()?;
    }
    Ok(())
}

const USEFUL_RAKEPACKS_TO_PULL_FROM_FEED: [&'static str; 2] = [".icons.pack", ".pages.pack"];

fn find_new_packs(feed_url: Url, processed_list_path: &Path) -> anyhow::Result<BTreeSet<String>> {
    let processed_file = OpenOptions::new()
        .read(true)
        .create(true)
        .open(processed_list_path)?;
    let br = BufReader::new(processed_file);
    let processed: Result<BTreeSet<String>, _> = br.lines().collect();
    let processed = processed.context("failed to read local processed list")?;

    let mut unprocessed: BTreeSet<String> = BTreeSet::new();

    let feed_lines = BufReader::new(reqwest::blocking::get(feed_url)?.error_for_status()?).lines();
    for line in feed_lines {
        let line = line?;
        if line.is_empty() {
            continue;
        }
        if !USEFUL_RAKEPACKS_TO_PULL_FROM_FEED
            .iter()
            .any(|ext| line.ends_with(ext))
        {
            // not a sort of rakepack we care about
            continue;
        }
        if processed.contains(&line) {
            continue;
        }
        unprocessed.insert(line.to_owned());
    }

    Ok(unprocessed)
}

pub fn handle_pack(
    pack: &Path,
    indexer_backend: &mut Box<dyn Backend>,
    seed_lookup: &SeedLookupTable,
    icon_store: &IconStore,
) -> anyhow::Result<()> {
    blue!("Indexing: ");
    yellow_ln!("{:?}", pack);

    let file = File::open(&pack)?;
    let decompressor = zstd::stream::Decoder::new(file)?;
    // TODO the decompressor has a buffer already, but we need this to see the end
    let mut buf_reader = BufReader::new(decompressor);
    let schema: String =
        serde_bare::from_reader(&mut buf_reader).context("failed to read schema ver")?;

    match schema.as_ref() {
        SCHEMA_RAKED_PAGES => {
            // TODO(unstable): this condition is `.has_data_left()` but it's unstable.
            while buf_reader.fill_buf().map(|b| !b.is_empty())? {
                handle_page_pack(&mut buf_reader, &seed_lookup, indexer_backend)
                    .context("failed to handle page pack")?;
            }
        }
        SCHEMA_RAKED_ICONS => {
            // TODO(unstable): this condition is `.has_data_left()` but it's unstable.
            while buf_reader.fill_buf().map(|b| !b.is_empty())? {
                handle_icon_pack(&mut buf_reader, &icon_store)
                    .context("failed to handle icon pack")?;
            }
        }
        _ => {
            bail!(
                "Wrong schema version: wanted e.g. {:?}, got {:?}",
                SCHEMA_RAKED_PAGES,
                &schema
            );
        }
    }
    Ok(())
}

pub fn handle_page_pack(
    buf_reader: &mut impl BufRead,
    seed_lookup: &SeedLookupTable,
    indexer_backend: &mut Box<dyn Backend>,
) -> anyhow::Result<()> {
    let page_record: PackRecord<RakedPageEntry> = serde_bare::from_reader(buf_reader)?;

    let document = page_record.record.document;

    let article_body = DenseTree::generate_textual_format(&document.body_content, false);
    let nonarticle_body = DenseTree::generate_textual_format(&document.body_remainder, false);

    let mut tags = seed_lookup
        .look_up(&Url::parse(page_record.url.as_ref())?)?
        .map(|seed: &Seed| seed.tags.iter().map(|cs| cs.to_string()).collect())
        .unwrap_or_else(|| Vec::with_capacity(0));

    // TODO(temporary): add antifeatures as tags, for now.
    let antifeatures = page_record.record.analysed_antifeatures;
    if antifeatures.contains(AnalysisAntifeatures::CLOUDFLARE) {
        tags.push("ZCloudflare".to_owned());
    }
    if antifeatures.contains(AnalysisAntifeatures::ADVERTS) {
        tags.push("ZAdverts".to_owned());
    }
    if antifeatures.contains(AnalysisAntifeatures::ANNOYANCE) {
        tags.push("ZAnnoyance".to_owned());
    }
    if antifeatures.contains(AnalysisAntifeatures::COOKIE_NAG) {
        tags.push("ZCookieNags".to_owned());
    }
    if antifeatures.contains(AnalysisAntifeatures::PRIVACY) {
        tags.push("ZTrackers".to_owned());
    }
    if antifeatures.contains(AnalysisAntifeatures::SOCIAL) {
        tags.push("ZSocialTrackers".to_owned());
    }

    // TODO Store the actual structure of the document in the store?

    let favicon_url_relative = document.head.effective_favicon_url();
    let favicon_url = Url::parse(page_record.url.as_ref())?.join(favicon_url_relative)?;
    let favicon_url = favicon_url.as_str();
    let mut favicon_url_hash_long = [0u8; 8];
    favicon_url_hash_long.copy_from_slice(&blake3::hash(favicon_url.as_bytes()).as_bytes()[0..8]);
    let favicon_url_hash = u64::from_le_bytes(favicon_url_hash_long);

    indexer_backend
        .add_document(BackendIndependentDocument {
            title: document.head.title,
            article_body,
            nonarticle_body,
            // TODO populate tags & antifeatures
            tags,
            url: page_record.url.to_string(),
            favicon_url_hash,
        })
        .context("failed to add document to indexer backend")?;

    Ok(())
}

pub fn handle_icon_pack(
    buf_reader: &mut impl BufRead,
    icon_store: &IconStore,
) -> anyhow::Result<()> {
    let page_record: PackRecord<IconEntry> = serde_bare::from_reader(buf_reader)?;

    let webp_bytes = page_record.record.webp_bytes;

    let mut favicon_url_hash_long = [0u8; 8];
    favicon_url_hash_long
        .copy_from_slice(&blake3::hash(page_record.url.as_ref().as_bytes()).as_bytes()[0..8]);
    let favicon_url_hash = u64::from_le_bytes(favicon_url_hash_long);

    icon_store.env.rw_txn(|txn| {
        icon_store
            .icons
            .put(txn, &favicon_url_hash.to_le_bytes(), &webp_bytes)
    })?;

    Ok(())
}

pub struct SeedLookupTable {
    pub by_prefix: PatriciaMap<Seed>,
    pub by_reduced_domain: HashMap<CompactString, Seed>,
}

pub async fn build_seed_lookup_table(
    mut seed_rx: Receiver<Seed>,
) -> anyhow::Result<SeedLookupTable> {
    let mut seed_lookup = SeedLookupTable {
        by_prefix: Default::default(),
        by_reduced_domain: HashMap::new(),
    };

    while let Some(seed) = seed_rx.recv().await {
        match &seed.url {
            UrlOrUrlPattern::Url(url_str) => {
                let url = Url::parse(url_str)?;
                let reduced_domain = get_reduced_domain(&url)
                    .with_context(|| format!("No domain in seed: '{url}'!"))?;
                seed_lookup
                    .by_reduced_domain
                    .insert(reduced_domain.into(), seed);
            }
            UrlOrUrlPattern::UrlPrefix(url_str) => {
                seed_lookup.by_prefix.insert(url_str.clone(), seed);
            }
        }
    }

    Ok(seed_lookup)
}

impl SeedLookupTable {
    pub fn look_up(&self, url: &Url) -> anyhow::Result<Option<&Seed>> {
        if let Some((_prefix, seed)) = self.by_prefix.get_longest_common_prefix(&url.as_str()) {
            return Ok(Some(seed));
        }

        let domain = get_reduced_domain(url)
            .with_context(|| format!("No domain in looked up URL: '{url}'"))?;
        if let Some(seed) = self.by_reduced_domain.get(domain.as_ref()) {
            return Ok(Some(seed));
        }

        Ok(None)
    }
}
