use anyhow::Context;
use quickpeep_index::backend::tantivy::TantivyBackend;
use quickpeep_index::backend::Backend;
use quickpeep_index::config::BackendConfig;
use serde::{Deserialize, Serialize};
use std::path::{Path, PathBuf};

#[derive(Serialize, Deserialize, Debug, Clone)]
/// Config for an indexer. All paths are relative to the config file if needed, but will be resolved
/// when loading.
pub struct IndexerConfig {
    /// Path to seeds
    pub seed_dir: PathBuf,

    pub index: IndexerOnlyConfig,
}

#[derive(Serialize, Deserialize, Debug, Clone)]
pub struct IndexerOnlyConfig {
    /// Path to the icon store
    pub icon_store: PathBuf,

    /// Configuration about which backend to use.
    pub backend: BackendConfig,
}

impl IndexerConfig {
    /// Loads a config at the specified path.
    /// Will resolve all the paths in the IndexerConfig for you.
    pub fn load(path: &Path) -> anyhow::Result<IndexerConfig> {
        let config_dir = path.parent().context("Can't get parent of config file.")?;
        let bytes = std::fs::read(path)?;
        let mut indexer_config: IndexerConfig = ron::de::from_bytes(&bytes)?;

        indexer_config.seed_dir = config_dir.join(indexer_config.seed_dir);
        indexer_config.index.icon_store = config_dir.join(indexer_config.index.icon_store);
        match &mut indexer_config.index.backend {
            BackendConfig::Tantivy(tantivy) => {
                tantivy.index_dir = config_dir.join(&tantivy.index_dir);
            }
            BackendConfig::Meili(_) => {}
        }

        Ok(indexer_config)
    }

    pub fn open_indexer_backend(&self) -> anyhow::Result<Box<dyn Backend>> {
        match &self.index.backend {
            BackendConfig::Tantivy(tantivy) => Ok(Box::new(
                TantivyBackend::open(&tantivy.index_dir)
                    .context("failed to open Tantivy backend")?,
            )),
            BackendConfig::Meili(_) => {
                todo!()
            }
        }
    }

    /// Returns the path to a text file which can be used for storing a list of processed rakepacks
    /// (needed for following rakepack streams over a network).
    pub fn processed_rakepack_path(&self) -> anyhow::Result<PathBuf> {
        match &self.index.backend {
            BackendConfig::Tantivy(tantivy) => {
                Ok(tantivy.index_dir.join("processed_rakepacks.lst"))
            }
            BackendConfig::Meili(_) => {
                todo!()
            }
        }
    }
}
