use adblock::lists::RuleTypes;
use anyhow::Context;
use chrono::{TimeZone, Utc};
use clap::Parser;
use colour::{blue, blue_ln, cyan_ln, green_ln, red_ln, yellow_ln};
use env_logger::Env;
use log::warn;
use quickpeep_raker::raking::analysis::{preload_adblock_engine, IpSet};
use quickpeep_raker::raking::page_extraction::PageExtractionService;
use quickpeep_raker::raking::references::references_from_urlrakes;
use quickpeep_raker::raking::{RakeIntent, RakeOutcome};
use quickpeep_raker::raking::{Raker, RAKER_USER_AGENT, TIME_LIMIT};
use quickpeep_structs::rake_entries::{AnalysisAntifeatures, RakedReference, ReferenceKind};
use reqwest::header::{HeaderMap, HeaderValue, USER_AGENT};
use reqwest::redirect::Policy;
use reqwest::Url;
use std::collections::BTreeSet;
use std::path::{Path, PathBuf};
use tokio::fs::File;

pub const ADBLOCK_FILTER_PATHS: [(AnalysisAntifeatures, &'static str); 4] = [
    (AnalysisAntifeatures::COOKIE_NAG, "cookie_nag"),
    (AnalysisAntifeatures::ANNOYANCE, "annoyance"),
    (AnalysisAntifeatures::PRIVACY, "privacy"),
    (AnalysisAntifeatures::ADVERTS, "adverts"),
];

/// Rakes one URL and prints out the description of it.
#[derive(Clone, Debug, Parser)]
pub struct Opts {
    url: Url,
    #[clap(short = 'i')]
    intent: Option<RakeIntent>,
}

#[tokio::main]
pub async fn main() -> anyhow::Result<()> {
    env_logger::Builder::from_env(Env::default().default_filter_or("info,quickpeep=debug")).init();

    let opts: Opts = Opts::parse();

    let mut header_map = HeaderMap::new();
    header_map.insert(USER_AGENT, HeaderValue::from_static(RAKER_USER_AGENT));

    let client = reqwest::ClientBuilder::new()
        .timeout(TIME_LIMIT)
        .default_headers(header_map.clone())
        // We want to handle redirects ourselves so we can track them...
        .redirect(Policy::none())
        .build()?;

    let redirect_following_client = reqwest::ClientBuilder::new()
        .timeout(TIME_LIMIT)
        .default_headers(header_map)
        // We want to handle redirects ourselves so we can track them...
        .redirect(Policy::limited(5))
        .build()?;

    let mut adblock_engines = Vec::new();

    for (antifeature, name) in &ADBLOCK_FILTER_PATHS {
        // TODO(not important) Don't hardcode these paths in quite as bad a way...
        let path = PathBuf::from(format!("./data/{}.adblock", name));
        if !path.exists() {
            warn!("Missing adblock rules: {:?}.", path);
            continue;
        }
        let file = File::open(&path).await?;
        adblock_engines.push((
            *antifeature,
            preload_adblock_engine(file, RuleTypes::All).await?,
        ));
    }

    let mut antifeature_ip_set = IpSet::new();

    let ips_file = File::open("./data/cf_ips.txt")
        .await
        .context("Failed to open CF IPs file")?;
    antifeature_ip_set.add_all_from_file(ips_file).await?;

    let raker = Raker {
        antifeature_ip_set,
        page_extraction: PageExtractionService::new(adblock_engines)?,
    };

    let intent = opts.intent.unwrap_or(RakeIntent::Any);
    let client_to_use = if intent == RakeIntent::Icon {
        &redirect_following_client
    } else {
        &client
    };

    let outcome = raker.rake(&opts.url, intent, client_to_use).await?;

    match outcome {
        RakeOutcome::RakedPage(page) => {
            let content_size = serde_bare::to_vec(&page)?.len();
            green_ln!("Page ({} bytes)", content_size);

            println!();
            let head = &page.page_entry.document.head;
            println!("Title: {}", head.title);
            println!("Language: {}", head.language);
            if !head.icon.is_empty() {
                println!("Icon: {}", head.icon);
            }
            match head.datetime {
                None => {
                    println!("Datetime: Unknown");
                }
                Some(datetime_unix) => {
                    let datetime = Utc.timestamp(datetime_unix, 0);
                    println!("Datetime: {}", datetime.format("%Y-%m-%d %H:%M:%S"));
                }
            }

            println!();
            print_references(&page.referrer_entry.references);
        }
        RakeOutcome::RakedFeed(feed) => {
            green_ln!("Feed");
            println!();
            let refs = references_from_urlrakes(&feed, ReferenceKind::FeedEntry);
            print_references(&refs);
        }
        RakeOutcome::RakedSitemap(sitemap) => {
            green_ln!("Sitemap");
            println!();
            let refs = references_from_urlrakes(&sitemap, ReferenceKind::SitemapEntry);
            print_references(&refs);
        }
        RakeOutcome::Redirect { reason, new_url } => {
            blue_ln!("Redirect ({:?})", reason);
            println!(" → {}", new_url.as_str());
        }
        RakeOutcome::TemporaryFailure(fail) => {
            yellow_ln!("Temporary Failure\n\t{:?}", &fail.reason);
        }
        RakeOutcome::PermanentFailure(fail) => {
            red_ln!("Permanent Failure\n\t{:?}", &fail.reason)
        }
        RakeOutcome::RakedIcon(icon) => {
            green_ln!("Icon");
            blue!("\tOriginal size: ");
            cyan_ln!("{} bytes", icon.original_size_in_bytes);
            blue!("\tPacked size: ");
            cyan_ln!("{} bytes", icon.webp_bytes.len());

            std::fs::write(Path::new("/tmp/qp-rake1.webp"), &icon.webp_bytes)?;
            println!("\n(saved as /tmp/qp-rake1.webp!)");
        }
    }

    Ok(())
}

fn print_references(refs: &BTreeSet<RakedReference>) {
    println!("{} References", refs.len());
    for reference in refs {
        println!("\t{:?} → {}", reference.kind, reference.target);
    }
}
