use anyhow::Context;
use clap::Parser;
use colour::{blue, yellow_ln};
use env_logger::Env;
use itertools::Itertools;
use quickpeep_raker::config;
use quickpeep_raker::storage::RakerStore;
use sqlx::query;
use std::collections::HashMap;
use std::path::PathBuf;

/// Seeds a raker's queue with URLs
#[derive(Clone, Debug, Parser)]
pub struct Opts {
    #[clap(long = "config")]
    config: Option<PathBuf>,
}

// TODO re-introduce refcounting
#[tokio::main]
pub async fn main() -> anyhow::Result<()> {
    env_logger::Builder::from_env(
        Env::default().default_filter_or("info,quickpeep_raker=debug,qp_raker=debug"),
    )
    .init();

    let opts: Opts = Opts::parse();

    let config_path = opts
        .config
        .unwrap_or_else(|| PathBuf::from("quickpeep.ron"));
    let config = config::RakerConfig::load(&config_path).context("Failed to load config")?;

    let store = RakerStore::open(&config.raker.database_uri).await?;

    let counts = count_on_hold(&store).await?;

    let counts = tokio::task::spawn_blocking(move || -> anyhow::Result<Vec<_>> {
        let sorted_counts = counts
            .into_iter()
            .map(|(string, count)| (count, string))
            .sorted()
            .rev()
            .collect_vec();

        Ok(sorted_counts)
    })
    .await??;

    blue!("№ URLs   ");
    yellow_ln!("Domain");
    for (count, string) in counts {
        println!("{:>6}   {}", count, string);
    }

    Ok(())
}

pub async fn count_on_hold(store: &RakerStore) -> anyhow::Result<HashMap<String, u32>> {
    store
        .ro_txn(move |txn| {
            Box::pin(async move {
                let rows = query!(
                    r#"
            SELECT d.name AS "domain_name", COUNT(1) AS "url_count" FROM urls u
            LEFT JOIN url_queue q ON u.url_id = q.url_id
            JOIN domains d USING (domain_id)
            WHERE u.visited_at_ts IS NULL AND q.url_id IS NULL
            GROUP BY d.name
        "#
                )
                .fetch_all(&mut *txn.txn)
                .await?;

                Ok(rows
                    .into_iter()
                    .map(|row| (row.domain_name, row.url_count.unwrap_or(0) as u32))
                    .collect::<HashMap<String, u32>>())
            })
        })
        .await
}
