use crate::rakepack_emitter::PackEmitterSettings;
use anyhow::Context;
use serde::{Deserialize, Serialize};
use std::net::SocketAddr;
use std::path::{Path, PathBuf};

#[derive(Serialize, Deserialize, Debug, Clone)]
/// Config for a raker. All paths are relative to the config file if needed, but will be resolved
/// when loading.
pub struct RakerConfig {
    /// Path to seeds
    pub seed_dir: PathBuf,

    pub raker: RakerOnlyConfig,
}

#[derive(Serialize, Deserialize, Debug, Clone)]
pub struct RakerOnlyConfig {
    /// Path to data files
    pub data_dir: PathBuf,

    /// URI to connect to Postgres.
    /// e.g. `postgres://user:password@host:port/dbname`
    pub database_uri: String,

    /// Directory where new rake packs will be emitted
    pub emit_dir: PathBuf,

    pub metrics: MetricsConfig,

    pub pack_emitter: PackEmitterSettings,

    pub rerake_timings: RerakeTimings,
}

#[derive(Serialize, Deserialize, Debug, Clone)]
pub struct RerakeTimings {
    /// How long, in days, between re-rakes of the same page?
    /// Suggested: 300
    pub page: i32,

    /// How long, in days, between re-rakes of feeds?
    /// Suggested: 10
    pub feed: i32,

    /// How long, in days, between re-rakes of icons?
    /// Suggested: 365
    pub icon: i32,
}

pub const DAY_SEC: i32 = 86400;

impl RakerConfig {
    /// Loads a config at the specified path.
    /// Will resolve all the paths in the RakerConfig for you.
    pub fn load(path: &Path) -> anyhow::Result<RakerConfig> {
        let config_dir = path.parent().context("Can't get parent of config file.")?;
        let bytes = std::fs::read(path)?;
        let mut raker_config: RakerConfig = ron::de::from_bytes(&bytes)?;

        raker_config.raker.data_dir = config_dir.join(raker_config.raker.data_dir);
        raker_config.seed_dir = config_dir.join(raker_config.seed_dir);
        raker_config.raker.emit_dir = config_dir.join(raker_config.raker.emit_dir);

        Ok(raker_config)
    }
}

#[derive(Serialize, Deserialize, Debug, Clone)]
pub struct MetricsConfig {
    #[serde(default)]
    pub bare_metrics: bool,
    #[serde(default)]
    pub prometheus: Option<SocketAddr>,
}
