use adblock::engine::Engine;
use adblock::lists::{ParseOptions, RuleTypes};
use anyhow::Context;
use html5ever::QualName;
use ipnetwork::IpNetwork;
use kuchiki::{ElementData, NodeData, NodeRef};
use lingua::Language;
use log::{debug, trace};
use reqwest::Url;
use std::collections::{BTreeSet, HashMap, HashSet};
use std::net::IpAddr;
use std::ops::Deref;
use tokio::io::{AsyncBufReadExt, AsyncRead, BufReader};

pub struct PreloadedEngine {
    rules: Vec<String>,
    rule_types: RuleTypes,
}

impl PreloadedEngine {
    pub fn build(&self) -> Engine {
        Engine::from_rules(
            &self.rules,
            ParseOptions {
                format: Default::default(),
                include_redirect_urls: false,
                rule_types: self.rule_types,
            },
        )
    }
}

/// 'Pre-load' the adblock engine.
/// This doesn't construct the adblock engine yet because the adblock engine is not `Send`.
pub async fn preload_adblock_engine<R: AsyncRead + Unpin>(
    reader: R,
    rule_types: RuleTypes,
) -> anyhow::Result<PreloadedEngine> {
    let mut br = BufReader::new(reader);
    let mut rules = Vec::new();
    let mut buf = String::new();
    loop {
        buf.clear();
        if br.read_line(&mut buf).await? == 0 {
            break;
        }
        rules.push(buf.trim().to_owned());
    }
    Ok(PreloadedEngine { rules, rule_types })
}

// Relevant:
// https://github.com/brave/adblock-rust/issues/152#issuecomment-771259069

pub struct ExtractedClassesAndIds {
    classes: Vec<String>,
    ids: Vec<String>,
}

pub fn extract_classes_and_ids_from_page(root: &NodeRef) -> ExtractedClassesAndIds {
    let mut class_set = HashSet::new();
    let mut id_set = HashSet::new();

    for node in root.inclusive_descendants() {
        if let Some(element) = node.0.as_element() {
            let attrs = element.attributes.borrow();
            if let Some(id) = attrs.get("id") {
                id_set.insert(id.to_owned());
            }
            if let Some(classes) = attrs.get("class") {
                for class in classes.trim().split_whitespace() {
                    class_set.insert(class.to_owned());
                }
            }
        }
    }

    ExtractedClassesAndIds {
        classes: class_set.into_iter().collect(),
        ids: id_set.into_iter().collect(),
    }
}

pub fn analyse_with_ad_block_cosmetic_filter(
    root: &NodeRef,
    engine: &Engine,
    url: &str,
    remove: bool,
) -> anyhow::Result<bool> {
    let mut matches = 0;

    let url_resources = engine.url_cosmetic_resources(url);
    let specialist_hide_selectors = if !url_resources.generichide {
        let ExtractedClassesAndIds { classes, ids } = extract_classes_and_ids_from_page(root);

        //eprintln!("ID {:#?}", ids);
        //eprintln!("CC {:#?}", classes);

        engine.hidden_class_id_selectors(&classes, &ids, &url_resources.exceptions)
    } else {
        Vec::with_capacity(0)
    };

    //eprintln!("UR {:#?}", url_resources);
    //eprintln!("sHS {:#?}", specialist_hide_selectors);
    //eprintln!("----");

    for rule in itertools::chain(specialist_hide_selectors, url_resources.hide_selectors) {
        if let Ok(result) = root.select(&rule) {
            for ele in result {
                trace!("Cosmetic Filter {:?} Matches {:?}", rule, ele);
                matches += 1;
                if remove {
                    ele.as_node().detach();
                }
            }
        } else {
            //eprintln!("(fail)");
        }
    }
    Ok(matches > 0)
}

/// Renames a node by fully replacing it in the tree.
fn replace_node(node: &NodeRef, name: Option<&str>) -> NodeRef {
    let ele = node.as_element().unwrap();
    let new_node = NodeRef::new(NodeData::Element(ElementData {
        name: QualName::new(
            None,
            ele.name.ns.clone(),
            name.unwrap_or(ele.name.local.deref()).into(),
        ),
        attributes: ele.attributes.clone(),
        template_contents: None,
    }));
    node.insert_after(new_node.clone());
    for child in node.children() {
        new_node.append(child);
    }
    node.detach();
    new_node
}

pub fn analyse_with_ad_block_network_filter(
    root: &NodeRef,
    engine: &Engine,
    url: &str,
) -> anyhow::Result<bool> {
    let mut matches = 0;

    let mut to_visit = vec![root.clone()];

    let this_url = Url::parse(url)?;

    // Element type → (Attribute, Request type)
    let processing_actions: HashMap<&'static str, (&'static str, &'static str)> = [
        ("a", ("href", "document")),
        ("img", ("src", "image")),
        ("script", ("src", "script")),
    ]
    .into_iter()
    .collect();

    while let Some(visit_next) = to_visit.pop() {
        for child in visit_next.children() {
            to_visit.push(child);
        }
        if let Some(element) = visit_next.as_element() {
            if let Some((attr_name, req_type)) = processing_actions.get(element.name.local.deref())
            {
                let attrs = element.attributes.borrow_mut();
                if let Some(href) = attrs.get(*attr_name) {
                    let should_remove = match this_url.join(href) {
                        Ok(href) => {
                            //eprintln!("check {:?}/{:?}/{:?}", href.as_str(), this_url.as_str(), *req_type);
                            engine
                                .check_network_urls(href.as_str(), this_url.as_str(), *req_type)
                                .matched
                        }
                        Err(err) => {
                            debug!("Removing erroneous {}: {:?} {:?}", attr_name, href, err);
                            true
                        }
                    };
                    drop(attrs);
                    if should_remove {
                        matches += 1;
                        replace_node(&visit_next, Some("disabled-node"));
                    }
                }
            }
        }
    }

    Ok(matches > 0)
}

pub fn guess_document_language(text: &str) -> Option<String> {
    let detector = lingua::LanguageDetectorBuilder::from_all_languages().build();
    detector
        .detect_language_of(text)
        .map(|lang: Language| lang.iso_code_639_1().to_string())
}

// TODO(perf) this isn't particularly efficient. Probably want a trie if it's important...
pub struct IpSet {
    ips: BTreeSet<IpNetwork>,
}

impl IpSet {
    pub fn new() -> IpSet {
        IpSet {
            ips: Default::default(),
        }
    }

    pub async fn add_all_from_file<R: AsyncRead + Unpin>(
        &mut self,
        reader: R,
    ) -> anyhow::Result<()> {
        let mut br = BufReader::new(reader);

        let mut buf = String::new();
        loop {
            buf.clear();
            if br.read_line(&mut buf).await? == 0 {
                break;
            }

            let trimmed = buf.trim();

            if trimmed.is_empty() {
                continue;
            }

            let ip_net = trimmed
                .parse::<IpNetwork>()
                .context("Parsing CIDR IP range")?;
            self.add(ip_net);
        }

        Ok(())
    }

    pub fn add(&mut self, network: IpNetwork) {
        // We jump through a couple of hoops to make sure we store the lowest address in the network,
        // since we use that for sorting.
        self.ips
            .insert(IpNetwork::new(network.network(), network.prefix()).unwrap());
    }

    pub fn contains(&self, addr: IpAddr) -> bool {
        let prefix = if addr.is_ipv4() {
            32
        } else {
            assert!(addr.is_ipv6());
            128
        };
        let addr_as_net =
            IpNetwork::new(addr, prefix).expect("Conversion to IpNetwork should be correct");
        for ipnet in self.ips.range(..=addr_as_net).rev().next() {
            if ipnet.contains(addr) {
                return true;
            }
        }

        false
    }
}

#[cfg(test)]
mod test {
    use crate::raking::analysis::IpSet;
    use ipnetwork::IpNetwork;
    use std::net::IpAddr;
    use std::str::FromStr;

    #[test]
    pub fn test_ipset_contains() {
        let mut set = IpSet::new();
        set.add(IpNetwork::from_str("1.2.3.4/16").unwrap());
        set.add(IpNetwork::from_str("1.1.2.3/16").unwrap());
        set.add(IpNetwork::from_str("85.42.36.17/24").unwrap());

        assert!(set.contains(IpAddr::from_str("1.2.42.42").unwrap()));
        assert!(set.contains(IpAddr::from_str("85.42.36.14").unwrap()));
        assert!(!set.contains(IpAddr::from_str("85.42.37.14").unwrap()));
    }
}
