use crate::raking::analysis::{
    analyse_with_ad_block_cosmetic_filter, analyse_with_ad_block_network_filter,
    guess_document_language, PreloadedEngine,
};
use crate::raking::references::dissolve_links;
use crate::raking::{normalise_language, RedirectReason, FEED_LINK_MIME_TYPES};
use adblock::engine::Engine;
use anyhow::{bail, Context};
use chrono::{DateTime, Utc};
use html5ever::tendril::TendrilSink;
use itertools::Itertools;
use kuchiki::NodeRef;
use log::{debug, error, trace, warn};
use quickpeep_densedoc::{DenseDocument, DenseHead, DenseTree};
use quickpeep_html_charset_detection::sniff;
use quickpeep_structs::rake_entries::AnalysisAntifeatures;
use quickpeep_utils::lazy::Lazy;
use reqwest::header::HeaderMap;
use reqwest::Url;
use serde::{Deserialize, Serialize};
use std::borrow::Borrow;
use tokio::runtime;
use tokio::sync::mpsc::Sender;
use tokio::sync::{mpsc, oneshot};
use tokio::task::LocalSet;

/// As page extraction requires dealing with things are are not `Send`, we send those tasks off to
/// another thread.
pub struct PageExtractionService {
    submission: Sender<ExtractionTask>,
}

pub struct ExtractionTask {
    content: Vec<u8>,
    url: Url,
    headers: HeaderMap,
    is_cf: bool,
    response: oneshot::Sender<anyhow::Result<ExtractedPage>>,
}

impl PageExtractionService {
    pub async fn extract(
        &self,
        content: Vec<u8>,
        url: Url,
        headers: HeaderMap,
        is_cf: bool,
    ) -> anyhow::Result<ExtractedPage> {
        let (resp_tx, resp_rx) = oneshot::channel();
        if let Err(_) = self
            .submission
            .send(ExtractionTask {
                content,
                url,
                headers,
                is_cf,
                response: resp_tx,
            })
            .await
        {
            bail!("Page extraction service has shut down");
        }
        resp_rx.await?
    }
}

impl PageExtractionService {
    pub fn new(
        adblock_engines: Vec<(AnalysisAntifeatures, PreloadedEngine)>,
    ) -> anyhow::Result<PageExtractionService> {
        let (submission, mut recv) = mpsc::channel::<ExtractionTask>(16);

        let rt = runtime::Builder::new_current_thread()
            .enable_all()
            .build()?;

        std::thread::spawn(move || {
            let internal = PageExtractionServiceInternal {
                adblock_engines: adblock_engines
                    .into_iter()
                    .map(|(aa, engine)| (aa, engine.build()))
                    .collect(),
            };

            let local = LocalSet::new();

            local.spawn_local(async move {
                while let Some(new_task) = recv.recv().await {
                    let result = internal.extract_page(
                        new_task.content,
                        new_task.url,
                        new_task.headers,
                        new_task.is_cf,
                    );
                    if let Err(_) = new_task.response.send(result) {
                        warn!("Requester went away before could send extracted page");
                    }
                }
            });

            // This will return once all senders are dropped and all spawned tasks have returned.
            rt.block_on(local);
        });

        Ok(PageExtractionService { submission })
    }
}

struct PageExtractionServiceInternal {
    adblock_engines: Vec<(AnalysisAntifeatures, Engine)>,
}

impl PageExtractionServiceInternal {
    fn extract_page(
        &self,
        content_bytes: Vec<u8>,
        url: Url,
        headers: HeaderMap,
        is_cf: bool,
    ) -> anyhow::Result<ExtractedPage> {
        let encoding = sniff(
            &content_bytes,
            true,
            headers.get("content-type").map(|hv| hv.as_bytes()),
        );
        let (content_text, _actual_codec_used, replacements_made) = encoding.decode(&content_bytes);
        if replacements_made {
            warn!("Character replacements made!");
        }
        let root_node: NodeRef = kuchiki::parse_html().one(content_text.borrow());

        // See whether this page is at the canonical URL for the page.
        // If it's not, then we redirect the raker to the canonical URL.
        if let Ok(canonical_link_node) = root_node.select_first("head link[rel=canonical]") {
            if let Some(canonical_href) = canonical_link_node.attributes.borrow().get("href") {
                let canonical_url = url
                    .join(canonical_href)
                    .context("Failed to resolve or parse canonical URL")?;

                if &canonical_url != &url {
                    return Ok(ExtractedPage::Redirect {
                        reason: RedirectReason::NotCanonical,
                        new_url: canonical_url,
                    });
                }
            }
        }

        // Try and dig up the page's language.
        // First try <html lang=...> since this is the modern way, and potentially the most trustworthy...
        let mut language = None;

        if let Ok(html_node) = root_node.select_first("html") {
            if let Some(lang) = html_node.attributes.borrow().get("lang") {
                language = Some(lang.trim().to_string());
            }
        }

        let mut no_follow = false;
        let mut no_index = false;

        // Find any restrictions on indexing this page or following any links.
        if let Ok(robots_nodes) = root_node.select("meta[name=robots]") {
            for node in robots_nodes {
                if let Some(content) = node.attributes.borrow().get("content") {
                    for directive in content
                        .split(|c: char| c.is_whitespace() || c == ',')
                        .filter(|s| !s.is_empty())
                    {
                        let none = directive.eq_ignore_ascii_case("none");
                        no_follow |= directive.eq_ignore_ascii_case("nofollow") | none;
                        no_index |= directive.eq_ignore_ascii_case("noindex") | none;
                    }
                }
            }
        }

        if language.is_none() {
            // Next fallback: prefer the content-language header baked into the page itself
            if let Ok(meta_node) = root_node.select_first("meta[http-equiv=content-language]") {
                if let Some(lang) = meta_node.attributes.borrow().get("content") {
                    language = Some(lang.trim().to_string());
                }
            }
        }

        if language.is_none() {
            // Next fallback: prefer the content-language received as a header
            if let Some(lang) = headers.get("content-language") {
                language = Some(lang.to_str()?.to_owned());
            }
        }

        let mut antifeature_flags = AnalysisAntifeatures::empty();

        if is_cf {
            antifeature_flags |= AnalysisAntifeatures::CLOUDFLARE;
        }

        for (engine_antifeature_flag, adblock_engine) in &self.adblock_engines {
            match analyse_with_ad_block_cosmetic_filter(
                &root_node,
                &adblock_engine,
                url.as_str(),
                true,
            ) {
                Ok(cosmetic_filters_tripped) => {
                    debug!("?cosmetic filters tripped: {}", cosmetic_filters_tripped);
                    antifeature_flags |= *engine_antifeature_flag;
                }
                Err(err) => {
                    error!("Cosmetic Filter Err {:?}", err);
                }
            };

            match analyse_with_ad_block_network_filter(&root_node, &adblock_engine, url.as_str()) {
                Ok(network_filters_tripped) => {
                    debug!("?network filters tripped: {}", network_filters_tripped);
                    antifeature_flags |= *engine_antifeature_flag;
                }
                Err(err) => {
                    error!("Network Filter Err {:?}", err);
                }
            };
        }

        let dense_doc = DenseTree::from_body(root_node.clone());
        let dense_doc_text = Lazy::new(Box::new(|| {
            DenseTree::generate_textual_format(&dense_doc, true)
        }));
        //eprintln!("^^^^^\n{}\n^^^^^", *dense_doc_text);

        if language.is_none() {
            // Final fallback: guess the language
            language = guess_document_language(&*dense_doc_text);
        }

        // Try and enforce some consistency in the language code;
        // we want codes like en_US rather than en-us.
        if let Some(language) = language.as_mut() {
            normalise_language(language);
        }

        let mut title = "".to_owned();

        if let Ok(title_node) = root_node.select_first("head title") {
            title = title_node.text_contents();
        }

        let mut feeds = Vec::new();
        let mut icon = None;

        for link_node in root_node.select("head link").into_iter().flatten() {
            if let Some(rel) = link_node.attributes.borrow().get("rel") {
                let rels = rel.split_whitespace().collect_vec();
                if rels.contains(&"icon") {
                    // This is an icon
                    if let Some(href) = link_node.attributes.borrow().get("href") {
                        let icon_url = url
                            .join(href)
                            .context("Failed to resolve or parse canonical URL to icon")?;

                        icon = Some(icon_url);
                    }
                } else if rels.contains(&"alternate") {
                    if let Some(rel_type) = link_node.attributes.borrow().get("type") {
                        if FEED_LINK_MIME_TYPES.contains(rel_type) {
                            if let Some(href) = link_node.attributes.borrow().get("href") {
                                let feed_url = url
                                    .join(href)
                                    .context("Failed to resolve or parse canonical URL to feed")?;

                                feeds.push(feed_url);
                            }
                        }
                    }
                }
            }
        }

        let metadata = find_page_metadata(root_node.clone())?;

        // Capture a copy of the unreadable document before it's too late.
        let unreadable_document = DenseTree::from_body(root_node.clone());

        let mut readability =
            quickpeep_moz_readability::Readability::new_from_node(root_node.clone());
        if let Err(err) = readability.parse(url.as_str()) {
            debug!("Failed to analyse readability: {:?}", err);
        }

        if title.is_empty() && !readability.metadata.title().is_empty() {
            // Fall back to the readability-derived page title
            title = readability.metadata.title().to_owned();
        }

        let datetime = metadata
            .publication_date
            .or(metadata.creation_date)
            .map(|dt| dt.timestamp());

        let mut document = DenseDocument {
            head: DenseHead {
                title,
                language: language.unwrap_or(String::with_capacity(0)),
                icon: icon
                    .map(|url| url.as_str().to_owned())
                    .unwrap_or(String::with_capacity(0)),
                datetime,
            },
            body_content: Vec::with_capacity(0),
            body_remainder: Vec::with_capacity(0),
        };

        if let Some(article_node) = readability.article_node {
            document.body_remainder = DenseTree::from_body(root_node);
            document.body_content = DenseTree::from_body(article_node);
        } else {
            // In the event of failing to pick out any article, we don't pick any article out.
            document.body_content = Vec::with_capacity(0);
            document.body_remainder = DenseTree::from_body(root_node);
        }

        // Dissolve links to save space
        dissolve_links(&mut document.body_content);
        dissolve_links(&mut document.body_remainder);

        Ok(ExtractedPage::Success {
            unreadable_document,
            document,
            feeds,
            antifeature_flags,
            no_follow,
            no_index,
        })
    }
}

fn parse_json_ld_date(input: String) -> Option<DateTime<Utc>> {
    DateTime::parse_from_rfc3339(&input)
        .ok()
        .map(|dt| dt.with_timezone(&Utc))
}

pub fn find_page_metadata(root_node: NodeRef) -> anyhow::Result<PageMetadata> {
    let mut meta = PageMetadata {
        publication_date: None,
        creation_date: None,
    };

    match root_node.select("script[type='application/ld+json']") {
        Ok(ld_nodes) => {
            for ld_node in ld_nodes {
                if let Some(_element) = ld_node.as_node().as_element() {
                    let json_text = ld_node.text_contents();
                    match serde_json::de::from_str::<JsonLdMetadata>(json_text.as_str()) {
                        Ok(jld) => {
                            let date_published =
                                jld.date_published.map(parse_json_ld_date).flatten();
                            let date_created = jld.date_created.map(parse_json_ld_date).flatten();
                            if date_published.is_some() {
                                meta.publication_date = date_published;
                            }
                            if date_created.is_some() {
                                meta.creation_date = date_created;
                            }
                        }
                        Err(err) => {
                            trace!("Bad JSON-LD: {:?}", err);
                        }
                    }
                }
            }
            Ok(meta)
        }
        Err(()) => Ok(meta),
    }
}

pub enum ExtractedPage {
    Success {
        unreadable_document: Vec<DenseTree>,
        document: DenseDocument,
        feeds: Vec<Url>,
        antifeature_flags: AnalysisAntifeatures,
        no_follow: bool,
        no_index: bool,
    },
    Redirect {
        reason: RedirectReason,
        new_url: Url,
    },
}

/// Metadata that may be found in a JSON-LD snippet on the page.
#[derive(Debug, Clone, Serialize, Deserialize)]
pub struct JsonLdMetadata {
    #[serde(rename = "@context")]
    #[serde(default)]
    pub context: Option<String>,

    #[serde(rename = "@type")]
    #[serde(default)]
    pub entity_type: Option<String>,

    #[serde(rename = "datePublished")]
    #[serde(default)]
    pub date_published: Option<String>,

    #[serde(rename = "dateCreated")]
    #[serde(default)]
    pub date_created: Option<String>,
}

pub struct PageMetadata {
    publication_date: Option<DateTime<Utc>>,
    creation_date: Option<DateTime<Utc>>,
}
