use crate::raking::UrlRaked;
use log::debug;
use quickpeep_densedoc::DenseTree;
use quickpeep_structs::rake_entries::{RakedReference, ReferenceKind};
use quickpeep_utils::dates::date_to_quickpeep_days;
use reqwest::Url;
use std::collections::BTreeSet;

/// Supported schemes.
/// References in all other schemes will be ignored.
pub const SUPPORTED_SCHEMES: [&'static str; 2] = ["http", "https"];

pub fn find_references(
    doc: &Vec<DenseTree>,
    feeds: &Vec<Url>,
    page_url: &Url,
) -> BTreeSet<RakedReference> {
    let mut refs = BTreeSet::new();

    fn add_link_refs(tree: &Vec<DenseTree>, refs: &mut BTreeSet<RakedReference>, page_url: &Url) {
        for node in tree {
            match node {
                DenseTree::Heading1(children) => {
                    add_link_refs(children, refs, page_url);
                }
                DenseTree::Heading2(children) => {
                    add_link_refs(children, refs, page_url);
                }
                DenseTree::Heading3(children) => {
                    add_link_refs(children, refs, page_url);
                }
                DenseTree::Heading4(children) => {
                    add_link_refs(children, refs, page_url);
                }
                DenseTree::Heading5(children) => {
                    add_link_refs(children, refs, page_url);
                }
                DenseTree::Heading6(children) => {
                    add_link_refs(children, refs, page_url);
                }
                DenseTree::Link {
                    children,
                    href,
                    nofollow,
                } => {
                    if !nofollow {
                        if let Ok(full_url) = page_url.join(&href) {
                            if full_url.domain().is_none() {
                                // Skip URLs that don't have a domain after being made absolute.
                                // This also skips IP addresses: we probably don't want to bother
                                // indexing content from explicit IP addresses.
                                continue;
                            }
                            if SUPPORTED_SCHEMES.contains(&full_url.scheme()) {
                                refs.insert(RakedReference {
                                    target: clean_url(&full_url).to_string(),
                                    kind: ReferenceKind::Link,
                                    last_mod: None,
                                });
                            } else {
                                debug!(
                                    "ignoring reference {:?}: not a supported scheme",
                                    full_url.as_str()
                                );
                            }
                        } else {
                            debug!("Can't join {:?} + {:?} to get full URL", page_url, href);
                        }
                    }
                    add_link_refs(children, refs, page_url);
                }
                DenseTree::Image { .. } => {}
                DenseTree::Text(_) => {}
            }
        }
    }

    add_link_refs(&doc, &mut refs, &page_url);

    for feed in feeds {
        if feed.domain().is_none() {
            // same rationale as above.
            continue;
        }
        refs.insert(RakedReference {
            target: clean_url(feed).as_str().to_owned(),
            kind: ReferenceKind::HeaderLinkedFeed,
            last_mod: None,
        });
    }

    refs
}

pub fn references_from_urlrakes(
    input: &Vec<UrlRaked>,
    ref_kind: ReferenceKind,
) -> BTreeSet<RakedReference> {
    input
        .iter()
        .map(|url_raked| RakedReference {
            target: clean_url(&url_raked.url).to_string(),
            kind: ref_kind,
            last_mod: url_raked
                .last_changed
                .map(|datetime| date_to_quickpeep_days(&datetime.date()).ok())
                .flatten(),
        })
        .collect()
}

pub fn dissolve_links(dense_tree: &mut Vec<DenseTree>) {
    let mut idx = 0;
    while idx < dense_tree.len() {
        let is_link = match &mut dense_tree[idx] {
            DenseTree::Heading1(children) => {
                dissolve_links(children);
                false
            }
            DenseTree::Heading2(children) => {
                dissolve_links(children);
                false
            }
            DenseTree::Heading3(children) => {
                dissolve_links(children);
                false
            }
            DenseTree::Heading4(children) => {
                dissolve_links(children);
                false
            }
            DenseTree::Heading5(children) => {
                dissolve_links(children);
                false
            }
            DenseTree::Heading6(children) => {
                dissolve_links(children);
                false
            }
            DenseTree::Link { children, .. } => {
                dissolve_links(children);
                true
            }
            DenseTree::Image { .. } => false,
            DenseTree::Text(_) => false,
        };
        if is_link {
            match dense_tree.remove(idx) {
                DenseTree::Link { children, .. } => {
                    dense_tree.splice(idx..idx, children);
                }
                _ => {
                    panic!("Implementation bug: Link being dissolved is not a Link.")
                }
            }
        } else {
            idx += 1;
        }
    }
}

pub fn clean_url(url: &Url) -> Url {
    let mut url = url.clone();
    url.set_fragment(None);

    url
}

#[cfg(test)]
mod test {
    use crate::raking::references::clean_url;
    use reqwest::Url;
    use std::str::FromStr;

    #[test]
    pub fn test_clean_url() {
        assert_eq!(
            clean_url(&Url::from_str("https://example.org:443/blah#hahah").unwrap()).as_str(),
            "https://example.org/blah",
        );

        assert_eq!(
            clean_url(&Url::from_str("https://example.org").unwrap()).as_str(),
            "https://example.org/",
        );
    }
}
