use crate::raking::{RakeIntent, TemporaryFailure};
use crate::storage::records::{DomainRecord, QueueUrlRecord};
use anyhow::{bail, ensure, Context};
use futures_core::future::BoxFuture;
use log::warn;
use metrics::{describe_gauge, gauge, Unit};
use quickpeep_utils::dates::QUICKPEEP_EPOCH_ST;
use quickpeep_utils::urls::get_reduced_domain;
use rand::random;
use reqwest::Url;
use sqlx::postgres::PgPoolOptions;
use sqlx::types::Json;
use sqlx::{query, query_as, Executor, PgPool, Postgres, Transaction};
use std::collections::HashSet;
use std::sync::atomic::AtomicU64;
use std::sync::Arc;
use std::time::{Duration, SystemTime};
use tokio::sync::Mutex;

pub mod maintenance;
pub mod records;

/// Handle to the store.
/// Doesn't need wrapping in Arc because it already is.
#[derive(Clone)]
pub struct RakerStore {
    pub pool: Arc<PgPool>,
    pub metrics: Option<Arc<RakerStoreMetrics>>,
}

impl RakerStore {
    pub async fn open(db_uri: &str) -> anyhow::Result<RakerStore> {
        let pool = PgPoolOptions::new()
            .after_connect(|conn, _meta| {
                Box::pin(async move {
                    conn.execute("SET default_transaction_isolation TO 'serializable'")
                        .await?;
                    Ok(())
                })
            })
            .connect(db_uri)
            .await
            .context("Failed to connect to Postgres pool")?;

        sqlx::migrate!()
            .run(&pool)
            .await
            .context("Failed to apply database migrations")?;

        Ok(RakerStore {
            pool: Arc::new(pool),
            metrics: Some(Arc::new(Default::default())),
        })
    }

    pub async fn rw_txn<'a, F, R>(&'a self, f: F) -> anyhow::Result<R>
    where
        F: for<'c> FnOnce(RakerTxn<'c, '_, true>) -> BoxFuture<'c, anyhow::Result<R>>
            + Send
            + Sync
            + 'a,
        R: Send + 'static,
    {
        self.txn(f).await
    }

    pub async fn ro_txn<'a, F, R>(&'a self, f: F) -> anyhow::Result<R>
    where
        F: for<'c> FnOnce(RakerTxn<'c, '_, false>) -> BoxFuture<'c, anyhow::Result<R>>
            + Send
            + Sync
            + 'a,
        R: Send + 'static,
    {
        self.txn(f).await
    }

    async fn txn<'a, const RW: bool, F, R>(&'a self, f: F) -> anyhow::Result<R>
    where
        F: for<'c> FnOnce(RakerTxn<'c, '_, RW>) -> BoxFuture<'c, anyhow::Result<R>>
            + Send
            + Sync
            + 'a,
        R: Send + 'static,
    {
        fn should_retry_db_err(sqlx_err: &sqlx::Error, retry_num: i32) -> bool {
            if let sqlx::Error::Database(db_err) = sqlx_err {
                if db_err.code().map(|code| code == "40001").unwrap_or(false) {
                    // This is a serialisation fault, retry the transaction.
                    warn!("serialisation failure: {db_err:?}; retrying try №{retry_num}");
                    return true;
                }
            }
            false
        }

        fn should_retry_err(err: &anyhow::Error, retry_num: i32) -> bool {
            if let Some(sqlx_err) = err.downcast_ref::<sqlx::Error>() {
                should_retry_db_err(sqlx_err, retry_num)
            } else {
                false
            }
        }

        let mut retry_num = 0;
        //TODO loop {
        retry_num += 1;
        let mut txn = self.pool.begin().await.context("Failed to get txn")?;

        match f(RakerTxn { txn: &mut txn }).await {
            Ok(r) => {
                if let Err(err) = txn.commit().await {
                    if should_retry_db_err(&err, retry_num) {
                        tokio::time::sleep(Duration::from_millis(50)).await;
                        bail!("continue;"); // TODO
                    }
                }
                return Ok(r);
            }
            Err(err) => {
                if should_retry_err(&err, retry_num) {
                    tokio::time::sleep(Duration::from_millis(50)).await;
                    bail!("continue;"); // TODO
                }
                return Err(err);
            }
        }
        //}
    }
}

/// Optional struct to store metrics. Intended for exporting over Prometheus so we can see at a
/// glance what the shape of the datastore looks like.
#[derive(Default)]
pub struct RakerStoreMetrics {
    pub active_domains_count: AtomicU64,
    pub backoff_domains_count: AtomicU64,
    pub queued_url_count: AtomicU64,
}

pub struct RakerTxn<'a, 't, const RW: bool> {
    pub txn: &'a mut Transaction<'t, Postgres>,
}

impl<'a, 't> RakerTxn<'a, 't, true> {
    pub async fn domain_to_id(&mut self, domain: &str) -> anyhow::Result<i32> {
        let new_id_row = query!(
            r#"
            INSERT INTO domains
            (name, domain_record)
            VALUES ($1, '{"rakeable_path_prefixes": {}}')
            ON CONFLICT DO NOTHING
            RETURNING domain_id
            "#,
            &domain
        )
        .fetch_optional(&mut *self.txn)
        .await?;

        if let Some(new_row) = new_id_row {
            return Ok(new_row.domain_id);
        }

        let row = query!(
            r#"
            SELECT domain_id FROM domains WHERE name = $1
            "#,
            &domain
        )
        .fetch_one(&mut *self.txn)
        .await?;
        Ok(row.domain_id)
    }

    /// Inserts a domain into the active domain table,
    /// generating a raffle ticket (and inserting it too).
    ///
    /// No-op if the domain is already an active domain or being backed off from.
    pub async fn insert_active_domain_with_new_raffle_ticket(
        &mut self,
        new_domain: String,
    ) -> anyhow::Result<()> {
        let domain_id = self.domain_to_id(&new_domain).await?;
        self.insert_active_domain_id_with_new_raffle_ticket(domain_id)
            .await?;
        Ok(())
    }

    /// Inserts a domain into the active domain table,
    /// generating a raffle ticket (and inserting it too).
    ///
    /// No-op if the domain is already an active domain or being backed off from.
    pub async fn insert_active_domain_id_with_new_raffle_ticket(
        &mut self,
        domain_id: i32,
    ) -> anyhow::Result<()> {
        if query!(
            "SELECT 1 AS _meh FROM domain_backoffs WHERE domain_id = $1",
            domain_id
        )
        .fetch_optional(&mut *self.txn)
        .await?
        .is_some()
        {
            // nop: backoff active
            return Ok(());
        }

        loop {
            let random_raffle_ticket: i32 = random();
            let rows_affected = query!(
                r#"
                INSERT INTO active_domain_raffle (domain_id, raffle)
                VALUES ($1, $2)
                ON CONFLICT DO NOTHING
                "#,
                domain_id,
                random_raffle_ticket
            )
            .execute(&mut *self.txn)
            .await?
            .rows_affected();

            if rows_affected == 0 {
                let r = query!(
                    r#"
                    SELECT raffle FROM active_domain_raffle WHERE domain_id = $1
                    "#,
                    domain_id,
                )
                .fetch_optional(&mut *self.txn)
                .await?;
                if r.is_some() {
                    break;
                }
            } else {
                break;
            }
        }

        Ok(())
    }

    /// Removes a domain from the active domains table, deleting its raffle ticket from the raffle
    /// table.
    ///
    /// Returns true if a deletion took place, and false if it did not.
    pub async fn remove_active_domain(&mut self, domain: &str) -> anyhow::Result<bool> {
        let domain_id = self.domain_to_id(domain).await?;
        let rowcount = query!(
            r#"
            DELETE FROM active_domain_raffle
            WHERE domain_id = $1
            "#,
            domain_id
        )
        .execute(&mut *self.txn)
        .await?
        .rows_affected();

        Ok(rowcount > 0)
    }

    /// Marks a URL as visited and takes it out of the queue.
    pub async fn mark_url_as_visited(
        &mut self,
        domain: &str,
        url_str: &str,
        visited_on_ts: i32,
        rerake_at_ts: Option<i32>,
    ) -> anyhow::Result<()> {
        // url_id SERIAL PRIMARY KEY NOT NULL,
        //     domain_id INTEGER NOT NULL REFERENCES domains(domain_id),
        //     scheme urlscheme NOT NULL,
        //     path TEXT NOT NULL,
        //     intent rakeintent NOT NULL,
        //     visited_at_ts INTEGER,
        let domain_id = self.domain_to_id(domain).await?;
        // TODO are domains always the same as on
        query!(
            r#"
            INSERT INTO urls (domain_id, url, intent, visited_at_ts)
            VALUES ($1, $2, $3::text::rakeintent, $4)
            ON CONFLICT (domain_id, url)
            DO UPDATE SET
                visited_at_ts = EXCLUDED.visited_at_ts
            "#,
            domain_id,
            url_str,
            "Any",
            visited_on_ts
        )
        .execute(&mut *self.txn)
        .await?;

        if let Some(rerake_at) = rerake_at_ts {
            query!(
                r#"
                INSERT INTO url_queue (url_id, rake_after_ts)
                SELECT url_id, $3 FROM urls WHERE domain_id = $1 AND url = $2
            "#,
                domain_id,
                url_str,
                rerake_at
            )
            .execute(&mut *self.txn)
            .await?;
        } else {
            self.dequeue_url(domain, url_str).await?;
        }

        Ok(())
    }

    /// Takes a URL out of the queue.
    pub async fn dequeue_url(&mut self, domain: &str, url_str: &str) -> anyhow::Result<()> {
        let domain_id = self.domain_to_id(domain).await?;
        query!(
            r#"
            DELETE FROM url_queue
            WHERE url_id IN (SELECT url_id FROM urls WHERE domain_id = $1 AND url = $2)
            "#,
            domain_id,
            url_str
        )
        .execute(&mut *self.txn)
        .await?;

        Ok(())
    }

    pub async fn start_backing_off(
        &mut self,
        domain: &str,
        backoff_for: i32,
        failed_url: String,
        failure: TemporaryFailure,
    ) -> anyhow::Result<()> {
        let domain_id = self.domain_to_id(domain).await?;

        let reinstate_at = SystemTime::now()
            .duration_since(*QUICKPEEP_EPOCH_ST)?
            .as_secs() as i32
            + backoff_for;

        let reason = format!("{failed_url} {failure:?}");
        query!(
            r#"
            INSERT INTO domain_backoffs (domain_id, backoff_until_ts, backoff_sec, reason)
            VALUES ($1, $2, $3, $4)
            ON CONFLICT (domain_id) DO UPDATE SET
                backoff_until_ts = EXCLUDED.backoff_until_ts,
                backoff_sec = EXCLUDED.backoff_sec,
                reason = EXCLUDED.reason
            "#,
            domain_id,
            reinstate_at,
            backoff_for,
            reason
        )
        .execute(&mut *self.txn)
        .await?;

        ensure!(
            self.remove_active_domain(domain).await?,
            "Can't back off from domain that's not active"
        );

        Ok(())
    }

    /// Reinstates backing-off domains up to the specified time.
    /// Returns the time of the next reinstatement, if there is one.
    pub async fn reinstate_backoffs(
        &mut self,
        up_to_ts: SystemTime,
    ) -> anyhow::Result<Option<SystemTime>> {
        let reinstate_up_to = up_to_ts.duration_since(*QUICKPEEP_EPOCH_ST)?.as_secs() as i32;

        let domain_id_rows = query!(
            r#"
            SELECT domain_id FROM domain_backoffs WHERE backoff_until_ts <= $1
        "#,
            reinstate_up_to
        )
        .fetch_all(&mut *self.txn)
        .await?;

        query!(
            r#"
            DELETE FROM domain_backoffs WHERE backoff_until_ts <= $1
        "#,
            reinstate_up_to
        )
        .execute(&mut *self.txn)
        .await?;

        for domain_id_row in domain_id_rows {
            self.insert_active_domain_id_with_new_raffle_ticket(domain_id_row.domain_id)
                .await?;
        }

        Ok(query!(
            r#"
            SELECT backoff_until_ts FROM domain_backoffs ORDER BY backoff_until_ts LIMIT 1
        "#
        )
        .fetch_optional(&mut *self.txn)
        .await?
        .map(|row| *QUICKPEEP_EPOCH_ST + Duration::from_secs(row.backoff_until_ts as u64)))
    }

    /// Enqueues a URL.
    /// If `last_modified` is specified, then this is a no-op if the page has already been
    /// visited since then.
    /// If `last_modified` is not specified, then this is a no-op if the page has already
    /// been visited.
    ///
    /// Returns: true if it was enqueued, false if nothing changed.
    pub async fn enqueue_url(
        &mut self,
        url_str: &str,
        last_modified: Option<SystemTime>,
        intent: RakeIntent,
        on_hold: bool,
    ) -> anyhow::Result<bool> {
        let url = Url::parse(url_str)?;
        let url_domain = get_reduced_domain(&url)
            .with_context(|| format!("No domain for to-be-enqueued URL: '{url}'!"))?;
        let domain_id = self.domain_to_id(&url_domain).await?;

        // Don't enqueue if it hasn't been modified since the last visit
        if let Some(last_modified) = last_modified {
            let last_visited = query!(
                r#"
                SELECT visited_at_ts FROM urls
                WHERE domain_id = $1 AND url = $2
            "#,
                domain_id,
                url_str
            )
            .fetch_optional(&mut *self.txn)
            .await?;

            let last_visited = last_visited
                .map(|lv| lv.visited_at_ts)
                .flatten()
                .map(|lv| *QUICKPEEP_EPOCH_ST + Duration::from_secs(lv as u64));

            if let Some(last_visited) = last_visited {
                if last_visited >= last_modified {
                    return Ok(false);
                }
            }
        }

        let intent_str: &'static str = intent.into();
        let rowcount = query!(
            r#"
            INSERT INTO urls (domain_id, url, intent, visited_at_ts)
            VALUES ($1, $2, $3::text::rakeintent, NULL)
            ON CONFLICT DO NOTHING
        "#,
            domain_id,
            url_str,
            intent_str
        )
        .execute(&mut *self.txn)
        .await?
        .rows_affected();

        if !(last_modified.is_some() || (rowcount > 0)) {
            return Ok(false);
        }

        if !on_hold {
            let url_id = query!(
                r#"
                SELECT url_id FROM urls WHERE domain_id = $1 AND url = $2
            "#,
                domain_id,
                url_str
            )
            .fetch_one(&mut *self.txn)
            .await?
            .url_id;

            query!(
                r#"
                INSERT INTO url_queue (url_id, rake_after_ts)
                VALUES ($1, 0)
                ON CONFLICT DO NOTHING
            "#,
                url_id
            )
            .execute(&mut *self.txn)
            .await?;

            // Activate the domain if needed...
            self.insert_active_domain_id_with_new_raffle_ticket(domain_id)
                .await?;
        }

        Ok(true)
    }

    pub async fn put_domain_record(
        &mut self,
        domain: &str,
        domain_record: DomainRecord,
    ) -> anyhow::Result<()> {
        let drjson =
            serde_json::to_string(&domain_record).context("failed to serialise DomainRecord")?;
        query!(
            r#"
            INSERT INTO domains (name, domain_record) VALUES ($1, $2::text::jsonb)
            ON CONFLICT (name) DO UPDATE SET domain_record = EXCLUDED.domain_record
        "#,
            domain,
            drjson
        )
        .execute(&mut *self.txn)
        .await?;
        Ok(())
    }
}

/// Registers metrics for the datastore. Call this once at startup.
pub fn register_datastore_metrics() -> anyhow::Result<()> {
    // Registration is if you want a handle to the Gauge. We don't care for now so leave it be...
    // register_gauge!("db_entries");
    describe_gauge!(
        "db_entries",
        Unit::Count,
        "Number of entries in a specific database."
    );
    // register_gauge!("db_size_bytes");
    describe_gauge!(
        "db_size_bytes",
        Unit::Bytes,
        "Size in bytes of all the pages used by a specific database."
    );
    Ok(())
}

#[derive(Clone, Debug, Eq, PartialEq)]
pub enum RandomActiveDomainAcquisition {
    GotOne { domain: String },
    AllBusy,
    NoneLeft,
}

/// Read-only implementations (but can also be used on RW transactions)
impl<'a, 't, const RW: bool> RakerTxn<'a, 't, RW> {
    /// Chooses a domain that is not busy, then marks it as busy.
    pub async fn acquire_random_active_domain(
        &mut self,
        busy_domains: Arc<Mutex<HashSet<String>>>,
    ) -> anyhow::Result<RandomActiveDomainAcquisition> {
        let mut busy_domains = busy_domains.lock().await;

        let rand_key: i32 = random();

        let next_domain_row_opt = query!(
            r#"
            SELECT name
            FROM active_domain_raffle
            JOIN domains USING (domain_id)
            WHERE raffle >= $1
            ORDER BY raffle
            LIMIT 1
        "#,
            rand_key
        )
        .fetch_optional(&mut *self.txn)
        .await?;

        if let Some(next_domain_row) = next_domain_row_opt {
            if !busy_domains.contains(&next_domain_row.name) {
                ensure!(
                    busy_domains.insert(next_domain_row.name.clone()),
                    "Domain already present even though it was checked to be free"
                );
                return Ok(RandomActiveDomainAcquisition::GotOne {
                    domain: next_domain_row.name,
                });
            }
        }

        // If we hit this, then our raffle ticket didn't bring anything non-busy up.
        // Fall back to a simple mechanism.
        let row_limit = (busy_domains.len() + 1) as i64;
        let next_domain_rows = query!(
            r#"
            SELECT name
            FROM active_domain_raffle
            JOIN domains USING (domain_id)
            WHERE raffle >= $1
            ORDER BY raffle
            LIMIT $2
        "#,
            rand_key,
            row_limit
        )
        .fetch_all(&mut *self.txn)
        .await?;

        if next_domain_rows.is_empty() {
            return Ok(RandomActiveDomainAcquisition::NoneLeft);
        }

        for next_domain_row in next_domain_rows {
            if !busy_domains.contains(&next_domain_row.name) {
                ensure!(
                    busy_domains.insert(next_domain_row.name.clone()),
                    "Domain already present even though it was checked to be free"
                );
                return Ok(RandomActiveDomainAcquisition::GotOne {
                    domain: next_domain_row.name,
                });
            }
        }

        Ok(RandomActiveDomainAcquisition::AllBusy)
    }

    pub async fn choose_url_for_domain(
        &mut self,
        domain: &str,
    ) -> anyhow::Result<Option<(String, QueueUrlRecord)>> {
        // TODO(perf): consider denormalising domain_id into url_queue
        let now_ts = SystemTime::now()
            .duration_since(*QUICKPEEP_EPOCH_ST)
            .unwrap()
            .as_secs() as i32;
        let row_opt = query!(
            r#"
            SELECT url, intent AS "intent: RakeIntent" FROM domains d
            JOIN urls u USING (domain_id)
            JOIN url_queue USING (url_id)
            WHERE d.name = $1 AND rake_after_ts < $2
            ORDER BY rake_after_ts
            LIMIT 1
        "#,
            domain,
            now_ts
        )
        .fetch_optional(&mut *self.txn)
        .await?;

        Ok(row_opt.map(|row| (row.url, QueueUrlRecord { intent: row.intent })))
    }

    pub async fn get_domain_record(
        &mut self,
        domain: &str,
    ) -> anyhow::Result<Option<DomainRecord>> {
        struct Row {
            pub domain_record: Json<DomainRecord>,
        }

        let domain_row = query_as!(
            Row,
            r#"
            SELECT domain_record AS "domain_record: Json<DomainRecord>" FROM domains WHERE name = $1
        "#,
            domain
        )
        .fetch_optional(&mut *self.txn)
        .await?;

        Ok(domain_row.map(|row| {
            let Json(result) = row.domain_record;
            result
        }))
    }

    /// Emits metrics for the datastore. Call this occasionally.
    pub async fn emit_datastore_metrics(&mut self) -> anyhow::Result<()> {
        for table in [
            "domains",
            "urls",
            "url_queue",
            "active_domain_raffle",
            "domain_backoffs",
        ] {
            let size_bytes = query!(
                "SELECT pg_relation_size($1::text::regclass) AS size_bytes",
                table
            )
            .fetch_one(&mut *self.txn)
            .await?
            .size_bytes
            .unwrap_or(-1);

            let num_rows = query!(
                r#"
                SELECT n_live_tup
                FROM pg_stat_user_tables
                WHERE relname = $1
            "#,
                table
            )
            .fetch_one(&mut *self.txn)
            .await?
            .n_live_tup
            .unwrap_or(-1);

            gauge!("db_entries", num_rows as f64, "db" => table);
            gauge!("db_size_bytes", size_bytes as f64, "db" => table);
        }

        Ok(())
    }
}

#[cfg(test)]
pub mod test {
    use super::*;
    use crate::raking::TemporaryFailureReason;
    use std::collections::BTreeSet;

    #[tokio::test]
    async fn test_reinstate_multiple_domains() -> anyhow::Result<()> {
        let store = RakerStore::open(env!("RAKER_DATABASE_URI")).await?;

        store
            .rw_txn(move |mut txn| {
                Box::pin(async move {
                    txn.insert_active_domain_with_new_raffle_ticket("a.invalid".to_owned())
                        .await?;
                    txn.insert_active_domain_with_new_raffle_ticket("b.invalid".to_owned())
                        .await?;
                    Ok(())
                })
            })
            .await?;

        let now = SystemTime::now();

        store
            .rw_txn(move |mut txn| {
                Box::pin(async move {
                    txn.start_backing_off(
                        "a.invalid",
                        300,
                        "".to_owned(),
                        TemporaryFailure {
                            reason: TemporaryFailureReason::ExcruciatingCrawlDelay(1),
                            backoff_sec: 300,
                        },
                    )
                    .await?;
                    txn.start_backing_off(
                        "b.invalid",
                        300,
                        "".to_owned(),
                        TemporaryFailure {
                            reason: TemporaryFailureReason::ExcruciatingCrawlDelay(1),
                            backoff_sec: 300,
                        },
                    )
                    .await?;
                    Ok(())
                })
            })
            .await?;

        store
            .ro_txn(move |mut txn| {
                Box::pin(async move {
                    assert_eq!(
                        txn.acquire_random_active_domain(Default::default()).await?,
                        RandomActiveDomainAcquisition::NoneLeft
                    );
                    Ok(())
                })
            })
            .await?;

        store
            .rw_txn(move |mut txn| {
                Box::pin(async move {
                    txn.reinstate_backoffs(now + Duration::from_secs(600))
                        .await?;
                    Ok(())
                })
            })
            .await?;

        store
            .ro_txn(move |mut txn| {
                Box::pin(async move {
                    let busy = Default::default();

                    let acq1 = txn.acquire_random_active_domain(Arc::clone(&busy)).await?;
                    let acq2 = txn.acquire_random_active_domain(Arc::clone(&busy)).await?;

                    assert!(
                        matches!((acq1.clone(), acq2.clone()), (
                RandomActiveDomainAcquisition::GotOne {
                    domain: dom1,
                    ..
                },
                RandomActiveDomainAcquisition::GotOne {
                    domain: dom2,
                    ..
                }
            ) if vec![dom1.as_ref(), dom2.as_ref()].into_iter().collect::<BTreeSet<&str>>() == vec![
                "a.invalid", "b.invalid"
            ].into_iter().collect::<BTreeSet<&str>>()),
                        "{:#?}",
                        (acq1, acq2)
                    );
                    Ok(())
                })
            })
            .await?;

        Ok(())
    }
}
